#include "includes.h"
extern Aplicacao *aplicacao;

//// Construtores
//
Jogo::Jogo(){
	int i;
	char strTmp[100];

	// carrega as imagens do jogo
	this->imagensJogo.titulo_jogo = auxDIBImageLoad("img/titulo_jogo.dat");
	this->imagensJogo.level_score = auxDIBImageLoad("img/level_score.dat");
	this->imagensJogo.normal = auxDIBImageLoad("img/normal.dat");
	this->imagensJogo.hard = auxDIBImageLoad("img/hard.dat");
	for(i=0;i<10;i++){
		sprintf(strTmp, "img/%d.dat", i);
		this->imagensJogo.number[i] = auxDIBImageLoad(strTmp);
	}
	this->imagensJogo.ring = auxDIBImageLoad("img/ring.dat");
	this->imagensJogo.pause = auxDIBImageLoad("img/pause.dat");
	//--
	this->estadoDoJogo.snakeWorld.snake = NULL;
	this->initJogo();
}
Jogo::~Jogo(){}

//// Mtodos Publicos
//
void Jogo::initJogo(DIFICULDADE nd){
	// Dados do Jogo
	//
	this->estadoDoJogo.nivel_dificuldade = nd; // por omissao
	this->estadoDoJogo.nomeJogador[0] = '\0';
	this->estadoDoJogo.pontuacao = 0;
	this->initPilhaAccao();
	this->estadoDoJogo.gamePaused = false;
	
	// Representaao interna do mundo da Cobra
	//
	this->estadoDoJogo.snakeWorld.alimento.x = 0;
	this->estadoDoJogo.snakeWorld.alimento.y = 0;
	this->estadoDoJogo.snakeWorld.tipo_alimento = FRACO;
	this->estadoDoJogo.snakeWorld.direccaoSnake = OESTE;
	this->estadoDoJogo.snakeWorld.tamanhoSnake = TAMANHO_SNAKE_INICIAL;
	this->estadoDoJogo.snakeWorld.growSnakeSize = 0;
	if(this->estadoDoJogo.snakeWorld.snake != NULL)
		free(this->estadoDoJogo.snakeWorld.snake);
	this->estadoDoJogo.snakeWorld.snake = (PONTO_2D *)malloc(sizeof(PONTO_2D) * TAMANHO_SNAKE_INICIAL);
	this->initSnake();
	this->geraAlimento();

	// fps
	if(this->estadoDoJogo.nivel_dificuldade == NORMAL)
		aplicacao->fps = FPS_NORMAL;
	else
		aplicacao->fps = FPS_HARD;
}

void Jogo::display_jogo(){
	this->drawTabuleiro();
	this->drawAlimentos();
	this->drawSnake();
}

void Jogo::interacao_jogo(unsigned char keyNormal, int keySpecial, int x, int y){
	if(keySpecial == -1){
			// Normal KeyBoard
		switch(keyNormal){
			case 'p':
				this->estadoDoJogo.gamePaused = ! this->estadoDoJogo.gamePaused;
				break;
			case 27: // Escape
					aplicacao->setEstadoApp(ESTADO_GAMEOVER);
					glutPostRedisplay();
					break;
			default:
					break;
		}
	} else {
		// Special KeyBoard
		switch(keySpecial){
			case GLUT_KEY_UP:
				if(! this->estadoDoJogo.gamePaused) this->pushAccao(UP);
				break;
			case GLUT_KEY_DOWN:
				if(! this->estadoDoJogo.gamePaused) this->pushAccao(DOWN);
				break;
			case GLUT_KEY_LEFT:
				if(! this->estadoDoJogo.gamePaused) this->pushAccao(LEFT);
				break;
			case GLUT_KEY_RIGHT:
				if(! this->estadoDoJogo.gamePaused) this->pushAccao(RIGHT);
				break;
			default: return;
		}
	}
}

// Actualiza o Estado do Jogo
void Jogo::update_jogo(){
	if(! this->estadoDoJogo.gamePaused){
		// actualiza jogo
		this->updateSnake();

		switch(this->detectaColisao()){
			case ALIMENTO:
				this->geraAlimento();
				break;
			case PAREDE:
			case COBRA:
				aplicacao->setEstadoApp(ESTADO_GAMEOVER);
				glutPostRedisplay();
				break;
			default: // NO_COLISION
				break;
		}
	}
}

//// Mtodos Privados
//

void Jogo::drawTabuleiro(){
	int ww;
	ww = aplicacao->getWindowWidth();
	
	// Cor dos Contornos
	glColor3f(0.5, 0.5, 0.8);
	// Zona do Hud
	glBegin(GL_QUADS);
		glVertex2i(20, 20);
		glVertex2i(ww-19, 20);
		glVertex2i(ww-19, 80);
		glVertex2i(20, 80);
	glEnd();
	
	glLineWidth(5.0);
	// Tabuleiro
	glColor3f(0.5, 0.0, 0.0);
	glBegin(GL_LINE_LOOP);
		glVertex2i(20, 100);
		glVertex2i(20 + LARGURA_MUNDO, 100);
		glVertex2i(20 + LARGURA_MUNDO, 100 + ALTURA_MUNDO);
		glVertex2i(20, 100 + ALTURA_MUNDO);
	glEnd();

	// Desenha as imagens do Tabuleiro
		// titulo
	glRasterPos2f(0, 63);
	glDrawPixels(this->imagensJogo.titulo_jogo->sizeX, this->imagensJogo.titulo_jogo->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.titulo_jogo->data);
		// Level e Score
	glRasterPos2f(455, 89);
	glDrawPixels(this->imagensJogo.level_score->sizeX, this->imagensJogo.level_score->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.level_score->data);
		// normal vs hard
	glRasterPos2f(570, 75);
	if(this->estadoDoJogo.nivel_dificuldade == NORMAL)
		glDrawPixels(this->imagensJogo.normal->sizeX, this->imagensJogo.normal->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.normal->data);
	else
		glDrawPixels(this->imagensJogo.hard->sizeX, this->imagensJogo.hard->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.hard->data);
		// digitos do score
	unsigned long int score = this->estadoDoJogo.pontuacao;
	int vectDigitos[100], deslocamento, i=0;
	do {
		vectDigitos[i] = score % 10;
		score /= 10;
		i++;
	} while(score > 0);
	for(i--, deslocamento=0; i>=0; i--, deslocamento += 31){
		glRasterPos2f(810 + deslocamento, 80);
		glDrawPixels(this->imagensJogo.number[vectDigitos[i]]->sizeX, this->imagensJogo.number[vectDigitos[i]]->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.number[vectDigitos[i]]->data);
	}
		// desenha o pause game se o jogo tiver pausado :)
	if(this->estadoDoJogo.gamePaused){
		static int nframe = 0;
		static bool mostraPause = true;

		if(++nframe >= aplicacao->fps){
			nframe = 0;
			mostraPause = ! mostraPause;
		}
		if(mostraPause){
			glRasterPos2f(8, 90);
			glDrawPixels(this->imagensJogo.pause->sizeX, this->imagensJogo.pause->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.pause->data);
		}
	}
		// ring (imagem de fundo do tabuleiro
	glRasterPos2f(21, ALTURA_MUNDO + 100);
	glDrawPixels(this->imagensJogo.ring->sizeX, this->imagensJogo.ring->sizeY, GL_RGB, GL_UNSIGNED_BYTE, this->imagensJogo.ring->data);
}

void Jogo::drawAlimentos(){
	PONTO_2D alimento;
	int raio = GROSSURA_ALIMENTO / 2;

	this->copyPonto2D(&alimento, &this->estadoDoJogo.snakeWorld.alimento);

	if((alimento.x == 0) && (alimento.y == 0))
		return;

	// Cor do Alimento
	switch(this->estadoDoJogo.snakeWorld.tipo_alimento){
		case FRACO:
			glColor3f(0.0, 0.8, 0.0);
			break;
		case MEDIO:
			glColor3f(1.0, 1.0, 0.0);
			break;
		case BRUTO:
			glColor3f(1.0, 0.0, 0.0);
			break;
	}

	// desenha um circulo
	glBegin(GL_TRIANGLE_FAN);
		glVertex2f(alimento.x, alimento.y); // centro
		for(float ang = 0.0; ang<2*PI; ang += PI/6)
			glVertex2f(alimento.x + (cos(ang) * raio), alimento.y - (sin(ang) * raio));
		glVertex2f(alimento.x + raio, alimento.y); // fim
	glEnd();
}

void Jogo::drawSnake(){
	int riscas;
	int i;
	int raio = GROSSURA_COBRA/2;

	for(i=0, riscas=0; i<=this->estadoDoJogo.snakeWorld.tamanhoSnake-1; i++, riscas++){
		PONTO_2D elemento;
		bool headZone = false;

		this->copyPonto2D(&elemento, &this->estadoDoJogo.snakeWorld.snake[i]);
		
		// cor (cor distinta para a cabea)
		if(i==this->estadoDoJogo.snakeWorld.tamanhoSnake-1){ // headZone
			glColor3f(0.8, 0.3, 0.0);
			headZone = true;
		}
		else {
			if(i==0){
				// cor do rabo
				glColor3f(0.8, 0.3, 0.0);
			}
			else {
				if((riscas<3) && (i>0)){
					// cor da cobra
					glColor3f(0.5, 0.8, 0.5);
				}
				else { 
					// cor da lista
					glColor3f(0.4, 0.8, 0.6);
					riscas = 0;
				}
			}
		}
		// desenha um elemento da cobra (todo bonitinho)
		glBegin(GL_POLYGON);
			glVertex2f(elemento.x - raio/2, elemento.y - raio); // v0
			glVertex2f(elemento.x - raio, elemento.y - raio/2); // v1
			glVertex2f(elemento.x - raio, elemento.y + raio/2); // v2
			glVertex2f(elemento.x - raio/2, elemento.y + raio); // v3
			glVertex2f(elemento.x + raio/2, elemento.y + raio); // v4
			glVertex2f(elemento.x + raio, elemento.y + raio/2); // v5
			glVertex2f(elemento.x + raio, elemento.y - raio/2); // v6
			glVertex2f(elemento.x + raio/2, elemento.y - raio); // v7
		glEnd();

		if(headZone){ // Renhicoco da cabea
			glColor3f(1.0, 1.0, 1.0);

			glBegin(GL_POLYGON);
				glVertex2f(elemento.x, elemento.y - (raio/2)); // v0
				glVertex2f(elemento.x - (raio/2), elemento.y); // v1
				glVertex2f(elemento.x, elemento.y + (raio/2)); // v2
				glVertex2f(elemento.x + (raio/2), elemento.y); // v3
			glEnd();
		}
	}
}

void Jogo::initSnake(){
	int nElx = LARGURA_MUNDO / DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
	int nEly = ALTURA_MUNDO / DISTANCIA_ENTRE_ELEMENTOS_SNAKE;

	// cabea
	this->estadoDoJogo.snakeWorld.snake[TAMANHO_SNAKE_INICIAL-1].x = 20 + (GROSSURA_COBRA/2) + ((nElx/2) * (DISTANCIA_ENTRE_ELEMENTOS_SNAKE));
	this->estadoDoJogo.snakeWorld.snake[TAMANHO_SNAKE_INICIAL-1].y = 100 + (GROSSURA_COBRA/2) + ((nEly/2) * (DISTANCIA_ENTRE_ELEMENTOS_SNAKE));
	// resto
	for(int i=TAMANHO_SNAKE_INICIAL-2; i>=0; i--){
		this->estadoDoJogo.snakeWorld.snake[i].x = this->estadoDoJogo.snakeWorld.snake[i+1].x + DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
		this->estadoDoJogo.snakeWorld.snake[i].y = this->estadoDoJogo.snakeWorld.snake[i+1].y;
	}
}

void Jogo::updateSnake(){
	int i;
	PROXIMA_ACCAO proximaAccao;
	PONTO_2D proximaPosCabeca;

	// verifica qual a proxima aco da cobra na pilha de aces
	proximaAccao = this->popAccao();
	switch(proximaAccao){
		case UP:
		case DOWN:
			if((this->estadoDoJogo.snakeWorld.direccaoSnake == NORTE) || (this->estadoDoJogo.snakeWorld.direccaoSnake == SUL))
				proximaAccao = NA;
			break;
		case LEFT:
		case RIGHT:
			if((this->estadoDoJogo.snakeWorld.direccaoSnake == ESTE) || (this->estadoDoJogo.snakeWorld.direccaoSnake == OESTE))
				proximaAccao = NA;
			break;
		default:
			break;
	}
	if(proximaAccao == NA){
		switch(this->estadoDoJogo.snakeWorld.direccaoSnake){
			case NORTE: proximaAccao = UP; break;
			case SUL: proximaAccao = DOWN; break;
			case ESTE: proximaAccao = RIGHT; break;
			case OESTE: proximaAccao = LEFT; break;
		}
	}
	//verifica a posio da cobra
	this->copyPonto2D(&proximaPosCabeca, &this->estadoDoJogo.snakeWorld.snake[this->estadoDoJogo.snakeWorld.tamanhoSnake-1]);
	switch(proximaAccao){
		case UP:
			this->estadoDoJogo.snakeWorld.direccaoSnake = NORTE;
			proximaPosCabeca.y -= DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
			break;
		case DOWN:
			this->estadoDoJogo.snakeWorld.direccaoSnake = SUL;
			proximaPosCabeca.y += DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
			break;
		case LEFT:
			this->estadoDoJogo.snakeWorld.direccaoSnake = OESTE;
			proximaPosCabeca.x -= DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
			break;
		case RIGHT:
			this->estadoDoJogo.snakeWorld.direccaoSnake = ESTE;
			proximaPosCabeca.x += DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
			break;
	}

	// actualiza a posio da cobra
	if(this->estadoDoJogo.snakeWorld.growSnakeSize == 0){
		// resto
		for(i=0; i<this->estadoDoJogo.snakeWorld.tamanhoSnake-1; i++)
			this->copyPonto2D(&this->estadoDoJogo.snakeWorld.snake[i], &this->estadoDoJogo.snakeWorld.snake[i+1]);
		
		// cabea
		this->copyPonto2D(&this->estadoDoJogo.snakeWorld.snake[this->estadoDoJogo.snakeWorld.tamanhoSnake-1], &proximaPosCabeca);
	} else {
		this->estadoDoJogo.snakeWorld.snake = (PONTO_2D *)realloc(this->estadoDoJogo.snakeWorld.snake, sizeof(PONTO_2D) * (++this->estadoDoJogo.snakeWorld.tamanhoSnake));
		this->copyPonto2D(&this->estadoDoJogo.snakeWorld.snake[this->estadoDoJogo.snakeWorld.tamanhoSnake-1], &proximaPosCabeca);
		this->estadoDoJogo.snakeWorld.growSnakeSize--;
	}
}

void Jogo::geraAlimento(){
	// quantos alimentos cabem na horizontal e na vertical dependendo da largura e altura do mundo respectivamente
	static int nElx = LARGURA_MUNDO / DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
	static int nEly = ALTURA_MUNDO / DISTANCIA_ENTRE_ELEMENTOS_SNAKE;
	static bool firstTime = true;

	if(firstTime){
		srand((unsigned int)time(NULL));
		firstTime = false;
	}

	// Tipo de Alimento gerado
	int alimentoSorteado = rand() % (100 + 1);
	this->estadoDoJogo.snakeWorld.tipo_alimento = BRUTO;
	if(alimentoSorteado <= PROB_ALIMENTO_FRACO)
		this->estadoDoJogo.snakeWorld.tipo_alimento = FRACO;
	if((alimentoSorteado > PROB_ALIMENTO_FRACO) && (alimentoSorteado <= (PROB_ALIMENTO_FRACO + PROB_ALIMENTO_MEDIO)))
		this->estadoDoJogo.snakeWorld.tipo_alimento = MEDIO;

	// posio do alimento no mundo
	PONTO_2D posAlimento;
	
	do {
		posAlimento.x = (20 + (GROSSURA_COBRA/2)) +  (((rand() + 1) % (nElx + 1)) * DISTANCIA_ENTRE_ELEMENTOS_SNAKE);
		posAlimento.y = (100 + (GROSSURA_COBRA/2)) + (((rand() + 1) % (nEly + 1)) * DISTANCIA_ENTRE_ELEMENTOS_SNAKE);
	} while(! this->posAlimentoValida(posAlimento));

	this->copyPonto2D(&this->estadoDoJogo.snakeWorld.alimento, &posAlimento);
}

TIPO_COLISAO Jogo::detectaColisao(){
	PONTO_2D posCabeca, posAlimento;
	int i, j;
	
	this->copyPonto2D(&posCabeca, &this->estadoDoJogo.snakeWorld.snake[this->estadoDoJogo.snakeWorld.tamanhoSnake - 1]);
	this->copyPonto2D(&posAlimento, &this->estadoDoJogo.snakeWorld.alimento);

	if(this->pontos2DEquals(posCabeca, posAlimento)){
		// Colisao com um alimento (dinner time ! lol)
		switch(this->estadoDoJogo.snakeWorld.tipo_alimento){
			case FRACO:
				this->estadoDoJogo.snakeWorld.growSnakeSize += UNIDADES_A_CRESCER / 2;
				this->estadoDoJogo.pontuacao += PONTOS_POR_ALIMENTO / 2;
				break;
			case MEDIO:
				this->estadoDoJogo.snakeWorld.growSnakeSize += UNIDADES_A_CRESCER;
				this->estadoDoJogo.pontuacao += PONTOS_POR_ALIMENTO;
				break;
			case BRUTO:
				this->estadoDoJogo.snakeWorld.growSnakeSize += UNIDADES_A_CRESCER * 4;
				this->estadoDoJogo.pontuacao += PONTOS_POR_ALIMENTO * 4;
				break;
		}
		return ALIMENTO;
	}
	if((posCabeca.x < 0 + 20) || (posCabeca.x > LARGURA_MUNDO + 20) || (posCabeca.y < 0 + 100) || (posCabeca.y > ALTURA_MUNDO + 100)){
		// Colisao contra a parede
		return PAREDE;
	}

	for(i=this->estadoDoJogo.snakeWorld.tamanhoSnake-2, j=i+1; i>=0; i--){
		// Colisao contra a propria cobra
		if(this->pontos2DEquals(this->estadoDoJogo.snakeWorld.snake[j], this->estadoDoJogo.snakeWorld.snake[i]))
			return COBRA;
	}
	// caso nao haja colisao nenhuma
	return NO_COLISION;
}

void Jogo::initPilhaAccao(){
	for(int i=0; i<MAX_ACCOES_PENDENTES; i++)
		this->estadoDoJogo.proximaAccao[i] = NA;
}

void Jogo::pushAccao(PROXIMA_ACCAO a){
	int i;

	for(i=0; i<MAX_ACCOES_PENDENTES; i++){
		if(this->estadoDoJogo.proximaAccao[i] == NA){
			this->estadoDoJogo.proximaAccao[i] = a;
			break;
		}
	}
}

PROXIMA_ACCAO Jogo::popAccao(){
	if(this->estadoDoJogo.proximaAccao[0] == NA)
		return NA;

	PROXIMA_ACCAO accao;
	int i;

	accao = this->estadoDoJogo.proximaAccao[0];
	for(i=1; i<MAX_ACCOES_PENDENTES; i++)
		this->estadoDoJogo.proximaAccao[i-1] = this->estadoDoJogo.proximaAccao[i];
	this->estadoDoJogo.proximaAccao[i-1] = NA;
	return accao;
}

bool Jogo::posAlimentoValida(PONTO_2D posAlimento){
	int i;

	for(i=0; i<this->estadoDoJogo.snakeWorld.tamanhoSnake; i++){
		if(this->pontos2DEquals(this->estadoDoJogo.snakeWorld.snake[i], posAlimento))
			return false;
	}

	return true;
}

PONTO_2D *Jogo::copyPonto2D(PONTO_2D *destino, PONTO_2D *origem){
	if(origem != NULL){
		destino->x = origem->x;
		destino->y = origem->y;

		return destino;
	}
	else return NULL;
}

bool Jogo::pontos2DEquals(PONTO_2D p1, PONTO_2D p2){
	if((p1.x == p2.x) && (p1.y == p2.y))
		return true;
	return false;
}

//// Getters and Setters
//

void Jogo::setNivelDificuldade(DIFICULDADE nd){
	this->estadoDoJogo.nivel_dificuldade = nd;
}

DIFICULDADE Jogo::getNivelDificuldade(){
	return this->estadoDoJogo.nivel_dificuldade;
}

char *Jogo::getNomeJogador(){
	char *newNome = new char[TAM_NOME_JOGADOR];
	strcpy(newNome, this->estadoDoJogo.nomeJogador);
	return newNome; // um Garbage Collector aki dava jeito (lol) o prof. Antonio Leito tem razo. Viva o LISP e o JAVA
}

void Jogo::setNomeJogador(char *nome){
	strcpy(this->estadoDoJogo.nomeJogador, nome);
}

unsigned long int Jogo::getPontuacao(){
	return this->estadoDoJogo.pontuacao;
}